# Name :        Componentpunch
# Description : Cuts hole for component through front and back mesh faces if parallel and tidies up
#               If not parallel creates new face on front face. Component must have a set of edges that define hole on cut plane.
# Author :      Bill Wood  at wh.wood@ntlworld.com
# Usage :       Select component(s)	on faces and run
# Date :        April 2006 - May 2009
# Type :        Architecture tool
# History:        


require 'sketchup.rb'

class Componentpunch
# ===============================

def Componentpunch.getoutline(punch_component)
# -------------------------------

  ret = nil
  compedges = []
  punch_component.entities.each { |elem |
  	if (elem.typename == "Edge") then
  	  if (elem.start.position.z == 0) && (elem.end.position.z == 0) then
  	    compedges.push(elem)
  	  end
    end
  }
  
  if compedges.length > 0 then
  
   startedges = Array::new(compedges.length,-1)
   endedges = Array::new(compedges.length,-1)

   compedges.length.times { | i |
   startvert = compedges[i].start
   endvert = compedges[i].end
     
   compedges.length.times { | j |
     	  if i != j then
     	    if compedges[j].used_by? startvert then
     	        startedges[i] = j
     	    end
     	    
     	    if compedges[j].used_by? endvert then
     	        endedges[i] = j
     	    end
     	  end
     }
   }
   
   ret = []

   edgelist = []
   compedges.length.times { | i |
     if (startedges[i] != -1) && (endedges[i] != -1) then
       edgelist.push(compedges[i])
     end
   }

   vertlist = []
   edgelist.length.times { | i |
     vertlist.push([edgelist[i].start.position,edgelist[i].end.position])
   }
   
	 loopsets = []
	 curset = vertlist[0]
	 endpos = curset[1]
	 points = [curset[0]] # log start

   vlen = vertlist.length
	 vlen.times { | i |

      verts = []
      vertlist.each { | elem |
        if (elem[0] == endpos) || (elem[1] == endpos) then
          verts.push(elem)
        end
      }

	 	  vertlist.delete(curset)
	 	  
	 	  if (verts.length < 2) then
			  loopsets.push(points)
        if !vertlist.empty? then
			    curset = vertlist[0]
			    endpos = curset[1]
			    points = [curset[0]] # log start
			  end
	 	  else
	 	    points.push(endpos)
        if verts.member?(curset) then
          verts.delete(curset)
        end
        curset = verts[0]

        curset.each { | elem | 
        if ((elem.distance endpos) != 0) then
            endpos = elem
            break
          end
        }
	 	  end
	 }


   # status = Geom.point_to_polygon_2D point, points, true
   # true denotes point on border is inside
   if (loopsets.length > 1) then
	   loopsets.length.times { | i |
	     curloop =  loopsets[i]
	     curinsideset = false

	     loopsets.length.times { | j |
	     	 if (i != j) then
	         testloop = loopsets[j]
	         insideset = true
	         curloop.each { | vert |
	           if !(Geom.point_in_polygon_2D vert, testloop, true) then
	             insideset = false
	           end
	         }
	         
	         if insideset then
	           curinsideset = insideset
	         end
	       end
	     }
	     
	     if !curinsideset then
	       ret.push(curloop)
	     end
	   }
	   
	 else
	   ret = loopsets
   end

  end
  
  return ret
end
# -------------------------------getoutline

def Componentpunch.initcomppunch
# -------------------------------

	model = Sketchup.active_model
	ss = model.selection
	
	if ss.empty? 
	  UI.messagebox("No selection. Select component(s) in face please.")
	  return nil
	end
	
	punchcomps  = []
	
	ss.each do |selem|
	  if ((selem.kind_of? Sketchup::ComponentInstance) && ((face = selem.glued_to) != nil)) then
	    edgeloop = Componentpunch.getoutline(selem)
	    if !edgeloop.nil? then 
	      punchcomps.push([selem,face,edgeloop])
	    end
	  end
	end
	
	ss.clear

	if punchcomps.empty? 
	  UI.messagebox("No glued components with correct cutout selected - Aborting.")
	  return nil
	end

	punchcomps.each do |punchcomp|
	  Componentpunch.punchcomponent(punchcomp)
  end

end
# -------------------------------initcomppunch

def Componentpunch.punchcomponent(punch_component)
# -------------------------------
  comp = punch_component[0]
  face = punch_component[1]
  edgeloops = punch_component[2]
  
  comp.glued_to = nil
  
  edgeloops.each { | points |

		  trans = comp.transformation
		  points.length.times { | i |
		  	points[i] = points[i].transform trans
		  }
		
		  newface = face.parent.entities.add_face points
		  newfacenorm = newface.normal.reverse!
		
		  if newface then
		    faces = face.all_connected
		    pushdist = -1
		
		    faces.each { | exgface |
		    	if (exgface.kind_of? Sketchup::Face) then
		    	  if (exgface.normal == newfacenorm) then
		    	  
			    	  pointonplane = points[0].project_to_plane exgface.plane
			    	  planedist =  pointonplane.distance(points[0])
			    	  
			    	  if (pushdist == -1) then 
			    	    pushdist = planedist
			    	  else
			    	    if (planedist < pushdist) then
			    	      pushdist = planedist
			    	    end
			    	  end
		
		    	  end #exgface.normal
		    	  
		      end #kindof
		      
		    } # faces.each
		    
		  	if (pushdist != -1) then
			    status = newface.pushpull -pushdist, false
			  end
		  
		  end # newface
  
  } # edgeloops

end
# -------------------------------drawcomponent

end
# Class==========================Componentfollowme

if( not file_loaded?("componentpunch.rb") )
    UI.menu("Plugins").add_item("Component punch") { Componentpunch.initcomppunch }
end

file_loaded("componentpunch.rb")